/* 
 * xnotify -- pop up notes on multiple displays and receive responses
 *             Copyright (C) 1993, David Partain
 *       Department of Computer and Information Sciences
 *              University of Linkoping, Sweden
 *                  581 83 Linkoping, Sweden
 *                     davpa@ida.liu.se
 *
 * xnotify is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * callbacks.c:  Version @(#)callbacks.c	1.1 6/30/93
 */
#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Dialog.h>

#include "xnotify.h"

/* 
 * callbacks: all have
 *	Widget widget, XtPointer client_data, XtPointer call_data" 
 * as parameters
 */
void ackcmd();
void cancelcmd();
void clearcmd();
void dismisscmd();
void maybecmd();
void nocmd();
void respondcmd();
void sendrespondcmd();
void yescmd();

/* 
 * external functions used by this module 
 */
extern void check_quit( /* int which */ );
extern Widget build_response( /* void */ );
extern void centerPopupOverCursor( /* Widget widget */ );
extern int get_index( /* Display *thisdisplay */ );
extern void closeup( /* int index, int response */ );

/* 
 * external variables used by this module 
 */
extern int closed;
extern int connections;
extern int silent;
extern int useX;
extern Widget server_toplevel;
extern XtArgVal serverlabelfont;
extern XtArgVal serverbuttonfont;
extern Widget listw;
extern conn_info connection[NUM_CONNS];

/* ARGSUSED */
/**********************************************************************
 * close up the response window. this is called from both the
 * exit and dismiss window buttons, the former of which is more
 * demanding.  additionally, if you try to exit the program before all
 * connections are closed, this callback is added to the buttons 
 * confirming whether the user really wants to quit.
 */
void dismisscmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  int which_one = (int)client_data;
  Widget shellw;
  if (closed == connections)
    exit(0);
  switch (which_one) {
    case TRYQUIT:
      check_quit(DISMISSCMD);
      break;
    case DOQUIT:
      exit(0);
      break;
    case DONTQUIT:
      XtPopdown(XtParent(XtParent(widget)));
      break;
    case WHACKED:
      if ((shellw = build_response()) != (Widget)0)
	XtPopup(shellw, XtGrabNone);
      XtPopdown(XtParent(XtParent(widget)));
      break;
    default:
      if (!silent)
        (void)fprintf(stderr, "shouldn't be here.\n");
  }
}

/* ARGSUSED */
void respondcmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  int index = (int)client_data;
  Widget shellw, dialogw;
  Dimension width;

  if (connection[index].display == (Display *)0)
    (void)fprintf(stderr, "something wrong\n");
  
  shellw = XtVaCreatePopupShell("DialogShell", transientShellWidgetClass,
  				connection[index].toplevel,
				XtNallowShellResize, True,
				XtNborderWidth, 1,
				NULL);
  dialogw = XtVaCreateManagedWidget("DialogWidget", dialogWidgetClass,
				shellw,
				XtNlabel, (String)"Message to send:",
				XtNborderWidth, 1,
				XtNvalue, "",
				NULL);
  XawDialogAddButton(dialogw, "send", sendrespondcmd, (XtPointer)dialogw);
  XawDialogAddButton(dialogw, "cancel", cancelcmd, (XtPointer)shellw);
  XtVaGetValues(dialogw, XtNwidth, &width, NULL);
  XtVaSetValues(dialogw, XtNwidth, width + 80, NULL);
  centerPopupOverCursor(shellw);
  XtPopup(shellw, XtGrabNone);
}

/* ARGSUSED */
void sendrespondcmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  Widget dialogw = (Widget)client_data;
  Widget shellw, labelw, boxw, buttonw;
  String input;
  char response[MESS_LEN];
  int index = -1;

  if ((index = get_index(XtDisplay(dialogw))) == -1)
    (void)fprintf(stderr, "can't get an index for the dialog widget.\n");
  input = XawDialogGetValueString(dialogw);
  if (!useX) {
    (void)fprintf(stdout, "%s says:\n\n\"%s\"\n\n",
		          (index == -1) 
		            ? XDisplayString(XtDisplay(dialogw))
		            : connection[index].name,
		          input);
    XtPopdown(XtParent((Widget)client_data));
    closeup(index, RESPOND);
    return;
  }
  bzero(response, sizeof(response));
  (void)sprintf(response, "%s says:\n\n\"%s\"",
		          (index == -1) 
		            ? XDisplayString(XtDisplay(dialogw))
		            : connection[index].name,
		          input);
  XtPopdown(XtParent((Widget)client_data));
  shellw = XtVaCreatePopupShell("ResponseShell", transientShellWidgetClass,
  				server_toplevel,
				XtNallowShellResize, True,
				XtNborderWidth, 1,
				NULL);
  boxw = XtVaCreateManagedWidget("ResponseBox", boxWidgetClass,
				shellw, NULL);
  labelw = XtVaCreateManagedWidget("ResponseLabel", labelWidgetClass,
				boxw,
				XtNlabel, (String)response,
				NULL);
  buttonw = XtVaCreateManagedWidget("ResponseDismiss", commandWidgetClass,
				boxw,
				XtNlabel, (XtArgVal)"Dismiss",
				XtNshapeStyle, XmuShapeOval,
				NULL);
  XtAddCallback(buttonw, XtNcallback, cancelcmd, (XtPointer)shellw);
  if (serverlabelfont) 
    XtVaSetValues(labelw, XtNfont, serverlabelfont, 0);
  if (serverbuttonfont) 
    XtVaSetValues(buttonw, XtNfont, serverbuttonfont, 0);
  centerPopupOverCursor(shellw);
  XtPopup(shellw, XtGrabNone);
  closeup(index, RESPOND);
}

/* ARGSUSED */
void ackcmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  closeup((int)client_data, ACK);
}

/* ARGSUSED */
void nocmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  closeup((int)client_data, NO);
}

/* ARGSUSED */
void yescmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  closeup((int)client_data, YES);
}

/* ARGSUSED */
void maybecmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  closeup((int)client_data, MAYBE);
}

/* ARGSUSED */
void clearcmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  XawListUnhighlight(listw);
}

/* ARGSUSED */
void cancelcmd(widget, client_data, call_data)
  Widget  widget;
  XtPointer client_data, call_data;
{
  XtPopdown((Widget)client_data);
}

