/* 
 * xnotify -- pop up notes on multiple displays and receive responses
 *             Copyright (C) 1993, David Partain
 *       Department of Computer and Information Sciences
 *              University of Linkoping, Sweden
 *                  581 83 Linkoping, Sweden
 *                     davpa@ida.liu.se
 *
 * xnotify is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * xnotify.c:  Version @(#)xnotify.c	1.3 6/30/93
 *
 * Welcome to xnotify.  A full description of the program can be found
 * in the man page.  
 *
 * History:
 * 	My group at the university here has a coffee machine in my
 *	office, and we all use it.  When one person makes coffee, it's
 *	a good idea to let the others know that it's ready.  Initially,
 *	we had a little thing that rsh'ed to all the machines and popped 
 *	up a little window on whatever the machine was.  You can see
 *	where this is going...  I rewrote the coffee notification to
 *	pop up on multiple displays, then decided I wanted to be able
 *	to send any text I wanted, then decided I wanted my choice of
 *	buttons, and so on, and so on.
 *
 * Thanks:
 *	Thanks to the people who tested along the way, offered
 *	comments and suggestions, and sent bug reports:
 *		Bernt Nilsson
 *		Ralph R{\"o}nnquist
 *		Ken Key
 *		Reed Wade
 *		Steve Moulton
 */

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <sys/time.h>
#include <pwd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>

#include "xnotify.h"

/* 
 * functions exported by this module 
 */
Widget build_response( /* void */ );
void centerPopupOverCursor( /* Widget widget */ );
void check_quit( /* int which */ );
void closeup( /* int index, int response */ );
int get_index( /* Display *thisdisplay */ );

/* 
 * external functions used by this module
 */
/* callbacks.c */
extern void dismisscmd();
extern void respondcmd();
extern void sendrespondcmd();
extern void ackcmd();
extern void nocmd();
extern void yescmd();
extern void maybecmd();
extern void clearcmd();
extern void cancelcmd();

/*
 * functions internal to this module
 */
static void add_buttons( /* int index */ );
static Widget build_box( /* Widget shellw, char *sender, char *recipients, 
			    char *when, char *message, int i */ );
static void buildlist( /* int i, char **argv */ );
static Dimension button_width( /* Widget button, Dimension *width,
				  int *num_buttons */ );
static void clear_conn( /* int i */ );
static int FatalError( /* Display *disp */ );
static void get_message( /* char *message */ );
static void get_recipients( /* char *recipients */ );
static void get_sender( /* void */ );
static void init_args( /* int argc, char **argv */ );
static void myMainLoop( /* XtAppContext app */ );
static int open_server( /* void */ );
static void print_pendings( /* void */ );
static void printout( /* int index, char *message */ );
static void set_geometry( /* Widget shellw, int index */ );
static void set_width( /* void */ );
static void start_X_and_conns( /* int argc, char **argv */ );
static void time_sent( /* void */ );
static Widget top_and_shell( /* int i */ );

/*
 * data accessible outside this module (in callbacks.c)
 */
conn_info connection[NUM_CONNS];	/* connection information */
Widget listw;				/* the list widget */
Widget server_toplevel;			/* toplevel widget for server */
int silent = 0;				/* get no talkative output */
int useX = 1;				/* am using X by default */
XtArgVal serverlabelfont;		/* font */
XtArgVal serverbuttonfont;		/* font */
int connections = 0;			/* counter for opened connections */
int closed = 0;				/* counter for closed connections */

/*
 * static (internal) data
 */
static XtAppContext app_context;	/* the application context */
static char *list[NUM_CONNS];		/* to put in the list widget */
static Widget boxw, senderw,		/* visible widgets */
	      receiverw, datew, 
	      messagew;
static Widget ackw, yesw, now,		/* buttons (optional) */
	      maybew,respondw;
static XtArgVal buttonfont;		/* font used in lots of places */
static Display *server_display;		/* display of the "server" */

/* these are controlled by command line options */
static int yesbutton = 0;		/* buttons */
static int nobutton = 0;
static int maybebutton = 0;
static int ackbutton = 0;
static int respondbutton = 0;
static int nobeep = 0;			/* don't beep when i get responses */

static jmp_buf SavedEnv;		/* catch kill window events */

/* strings in the outgoing message */
static char sender[STRLEN];		/* sender's name */
static char *when = NULL;		/* represents time msg sent */

/* msg sent from respond button */
static char *response_mess = NULL;

/*  */
void main(argc, argv)
  int argc;
  char **argv;
{
  Widget shellw;			/* outer widgets */
  char message[MESS_LEN];		/* message to send */
  char recipients[STRLEN];		/* string holding recipients' ids */
  int i;

  init_args(argc, argv);
  time_sent();
  get_message(message);
  get_sender();
  start_X_and_conns(argc, argv);
  get_recipients(recipients);

  /* go through the displays and pop up a window with the message */
  for (i = 0; i < connections; i++) {
    buildlist(i,argv);
    if ((shellw = top_and_shell(i)) == (Widget) 0) {
      clear_conn(i);
      continue;		/* can't keep going on this connection */
    }

    set_geometry(shellw, i);
    boxw = build_box(shellw, sender, recipients, when, message, i);
    add_buttons(i);
    set_width();

    XtPopup(shellw, XtGrabNone);
    if (!nobeep)
      XBell(connection[i].display,100);
  }
  if (connections == closed) {
    (void)fprintf(stderr, "no connections made. exiting.\n");
    exit(0);
  }

  if (useX && open_server() && ((shellw = build_response()) != (Widget) 0)) {
    XtPopup(shellw, XtGrabNone);
  }
  myMainLoop(app_context);
}

/*  */
/**********************************************************************
 * this is here so that if there are no resources set for the
 * xnotify window, it won't sit there waiting for someone to place
 * the window before it moves on to the other connections.
 */
static void set_geometry(shellw, index)
  Widget shellw;
  int index;
{
  char *geo = NULL;
  XrmDatabase database = NULL;
  XrmValue geovalue;

#ifndef __hpux
  if (connection[index].display->xdefaults != NULL) {
    database = connection[index].display->db;
    XrmGetResource(database, "xnotify*Geometry", "XtCGeometry", 
		   &geo, &geovalue);
  }
#endif

  if (database == NULL || geo == (char *)0)
    XtVaSetValues(shellw, XtNgeometry, (String)"+380+20", NULL);
}

/*  */
/**********************************************************************
 * this function for getting and setting values is there entirely
 * for my aesthetic pleasure.  it's supposed to center things, make
 * all buttons combined as wide as the widest widget.  it
 * ain't perfect, but i like it okay.
 */
static void set_width() 
{
  Dimension width, largest;		/* width of widgets */
  Dimension yes_width, no_width, 
	ack_width, maybe_width,
	respond_width;			/* individual button widths */
  Dimension buttons_width = 0;		/* total button width */
  int num_buttons = 0;

  /* find the widest widget */
  largest = 0;
  XtVaGetValues(messagew, XtNwidth, &width, NULL);
  largest = MAX(largest, width);
  XtVaGetValues(datew, XtNwidth, &width, NULL);
  largest = MAX(largest, width);
  XtVaGetValues(senderw, XtNwidth, &width, NULL);
  largest = MAX(largest, width);
  XtVaGetValues(receiverw, XtNwidth, &width, NULL);
  largest = MAX(largest, width);

  /* get the combined width of the button widgets */
  buttons_width = 0;
  num_buttons = 0;
  /* this block figures out the total width of all buttons */
  if (ackbutton)
    buttons_width += button_width(ackw, &ack_width, &num_buttons);

  if (yesbutton)
    buttons_width += button_width(yesw, &yes_width, &num_buttons);

  if (nobutton)
    buttons_width += button_width(now, &no_width, &num_buttons);

  if (maybebutton)
    buttons_width += button_width(maybew, &maybe_width, &num_buttons);

  if (respondbutton)
    buttons_width += button_width(respondw, &respond_width, &num_buttons);

  /* resize stuff so that it looks good */
  if (buttons_width > largest)
    largest = buttons_width;
  else {
    /* need to resize the buttons */
    Dimension resize;
    resize = ((largest - buttons_width)/num_buttons);
    if (maybebutton)
      XtVaSetValues(maybew, XtNwidth, maybe_width + resize, 0);
    if (nobutton)
      XtVaSetValues(now, XtNwidth, no_width + resize, 0);
    if (yesbutton)
      XtVaSetValues(yesw, XtNwidth, yes_width + resize, 0);
    if (ackbutton)
      XtVaSetValues(ackw, XtNwidth, ack_width + resize, 0);
    if (respondbutton)
      XtVaSetValues(respondw, XtNwidth, respond_width + resize, 0);
  }
  
  /* make widgets wider so they'll be centered */
  XtVaSetValues(datew, XtNwidth, largest, 0);
  XtVaSetValues(messagew, XtNwidth, largest, 0);
  XtVaSetValues(receiverw, XtNwidth, largest, 0);
  XtVaSetValues(senderw, XtNwidth, largest, 0);
}

/*  */
/**********************************************************************
 * determine the width of a button.
 */
static Dimension button_width(button, width, num_buttons)
  Widget button;
  Dimension *width;
  int *num_buttons;
{
  Dimension local_width;

  if (buttonfont) 
    XtVaSetValues(button, XtNfont, buttonfont, 0);
  XtVaGetValues(button, XtNwidth, &local_width, NULL);
  *width = local_width;
  local_width += BW_AND_DD;
  *num_buttons += 1;
  return local_width;
}

/*  */
/**********************************************************************
 * add all the buttons that they called for on the command line.
 */
static void add_buttons(index)
  int index;
{
  if (yesbutton) {
    yesw = XtVaCreateManagedWidget("yes_widget", commandWidgetClass,
				boxw, 
				XtNlabel, (XtArgVal)"Yes",
				XtNshapeStyle, XmuShapeOval,
				NULL);
    XtAddCallback(yesw, XtNcallback, yescmd, (XtPointer)index);
  }
  if (nobutton) {
    now = XtVaCreateManagedWidget("no_widget", commandWidgetClass,
				boxw, 
				XtNlabel, (XtArgVal)"No",
				XtNshapeStyle, XmuShapeOval,
				NULL);
    XtAddCallback(now, XtNcallback, nocmd, (XtPointer)index);
  }
  if (maybebutton) {
    maybew = XtVaCreateManagedWidget("maybe_widget", commandWidgetClass,
				     boxw, 
				     XtNlabel, (XtArgVal)"Maybe",
				     XtNshapeStyle, XmuShapeOval,
				     NULL);
    XtAddCallback(maybew, XtNcallback, maybecmd, (XtPointer)index);
  }
  if (respondbutton) {
    respondw = XtVaCreateManagedWidget("respond_widget", commandWidgetClass,
				       boxw, 
				       XtNlabel, (XtArgVal)"Respond",
				       XtNshapeStyle, XmuShapeOval,
				       NULL);
    XtAddCallback(respondw, XtNcallback, respondcmd, (XtPointer)index);
  }

  if ((!yesbutton && !nobutton && !maybebutton && !respondbutton)
      || ackbutton) {
    ackbutton = 1;
    ackw = XtVaCreateManagedWidget("button_widget", commandWidgetClass,
				   boxw, 
				   XtNlabel, (XtArgVal)"Acknowledged",
				   XtNshapeStyle, XmuShapeOval,
				   NULL);
    XtAddCallback(ackw, XtNcallback, ackcmd, (XtPointer)index);
  }
}

/*  */
/**********************************************************************
 * build the box that they're going to see.
 */
static Widget build_box(shellw, sender, recipients, when, message, i)
  Widget shellw;
  char *sender;
  char *recipients;
  char *when;
  char *message;
  int i;
{
   XtArgVal timefont, labelfont;

   boxw = XtCreateManagedWidget("box_widget", boxWidgetClass,
				shellw, NULL, 0);

  senderw = XtVaCreateManagedWidget("sender_widget", labelWidgetClass,
				boxw, 
				XtNlabel, (XtArgVal) sender,
				XtNborderWidth, 0,
				NULL);
				
  receiverw = XtVaCreateManagedWidget("receive_widget", labelWidgetClass,
				boxw, 
				XtNlabel, (XtArgVal) recipients,
				XtNborderWidth, 0,
				NULL);
  datew = XtVaCreateManagedWidget("date_widget", labelWidgetClass,
				boxw, 
				XtNlabel, (XtArgVal)when,
				XtNborderWidth, 0,
				XtNjustify, XtJustifyCenter,
				NULL);

  messagew = XtVaCreateManagedWidget("label_widget", labelWidgetClass,
				boxw,
				XtNlabel, (XtArgVal) message,
				XtNborderWidth, 1,
				NULL);
/* 
 * reset fonts for widgets if we can find the one we want.  they
 * were created with defaults, but this way i can try to make it
 * prettier
 */
  if ((timefont = 
       (XtArgVal)XLoadQueryFont(connection[i].display,"8x13"))!=(XtArgVal)0)
    XtVaSetValues(datew, XtNfont, timefont, 0);
  if ((labelfont = 
       (XtArgVal)XLoadQueryFont(connection[i].display,"9x15"))!=(XtArgVal)0)
    XtVaSetValues(messagew, XtNfont, labelfont, 0);
  if ((buttonfont=
       (XtArgVal)XLoadQueryFont(connection[i].display,"6x9"))!=(XtArgVal)0){
    XtVaSetValues(senderw, XtNfont, buttonfont, 0);
    XtVaSetValues(receiverw, XtNfont, buttonfont, 0);
  }
  return boxw;
}

/*  */
/**********************************************************************
 * a copy of XtAppMainLoop so that i can jump back into it after fatal
 * errors.
 */
static void myMainLoop(app)
  XtAppContext app;
{
  XEvent event;
  for (;;) {
    /* i'll jump here if someone whacks a window */
    (void)setjmp(SavedEnv);
    XtAppNextEvent(app, &event);
    XtDispatchEvent(&event);
  }
}

/*  */
/**********************************************************************
 * change the response window so that it reflects what the person 
 * at a particular display said.
 */
void closeup(index, response)
  int index;
  int response;
{
  /* 
   * this is just to give a little delay between the
   * button push and the closing of the window.  kind of abrupt otherwise.
   */
  (void)sleep(1);
  closed++;
  switch (response) {
    case YES:
      printout(index,"says yes");
      break;
    case NO:
      printout(index,"says no");
      break;
    case MAYBE:
      printout(index,"says maybe");
      break;
    case ACK:
      printout(index,"acknowledges");
      break;
    case VIOLENT:
      printout(index,"whacked the window.  Scum");
      break;
    case RESPOND:
      printout(index,"responded through \"respond\"");
      break;
    default:
      if (!silent)
        (void)fprintf(stderr, "unknown value for response.\n");
      break;
  }

  if (useX)
    XawListChange(listw, list, connections + 1, 0, True);
  if (!nobeep)
    XBell(server_display,100);
  XtCloseDisplay(connection[index].display);
  if (response != VIOLENT)
    XtDestroyWidget(connection[index].toplevel);
  connection[index].display = NULL;
  connection[index].toplevel = NULL;
  if (!useX)
    if (connections == closed)
      exit(0);
}

/*  */
/**********************************************************************
 * this builds the response window.  NOTE that you have to call
 * open_server() before you do this, otherwise you'll get stuff on
 * stdout.
 */
Widget build_response()
{
  Widget shellw, formw, labelw, exitw;
  XtArgVal serverlistfont;
  static int done = 0;

  /* if the server stuff ain't there, we'd better assume noX */
  if ((!server_display) || (!server_toplevel)) {
    (void)fprintf(stderr, "no server display or toplevel. assuming noX\n");
    useX = 0;
    nobeep = 1;
    return (Widget)0;
  }

  shellw = XtVaCreatePopupShell("Response", applicationShellWidgetClass,
				server_toplevel, XtNallowShellResize, True,
				NULL);
  formw = XtVaCreateManagedWidget("ResponseForm", formWidgetClass,
				shellw, 
				XtNborderWidth, 0,
				NULL);
  if (!done) {
    list[connections] = (char *) malloc((unsigned) STRLEN);
    (void)sprintf(list[connections], "Time: %s", when);
    done = 1;
  }
  listw = XtVaCreateManagedWidget("ResponseList", listWidgetClass,
				formw,
				XtNresizable, True,
				XtNborderWidth, 0,
				XtNlist, list,
				XtNdefaultColumns, 1,
				XtNnumberStrings, (connections + 1),
				NULL);
  labelw = XtVaCreateManagedWidget("ResponseLabel", labelWidgetClass,
				formw,
				XtNlabel, (XtArgVal) response_mess,
				XtNfromVert, (Widget)listw,
				NULL);
  exitw = XtVaCreateManagedWidget("ResponseExit", commandWidgetClass,
				formw,
				XtNlabel, (XtArgVal)"Dismiss Window",
				XtNfromVert, (Widget)labelw,
				XtNshapeStyle, XmuShapeOval,
				NULL);
  XtAddCallback(listw, XtNcallback, clearcmd, (XtPointer)0);
  XtAddCallback(exitw, XtNcallback, dismisscmd, (XtPointer)TRYQUIT);

  if ((serverlistfont =
      (XtArgVal)XLoadQueryFont(server_display,"8x13")) != (XtArgVal)0)
    XtVaSetValues(listw, XtNfont, serverlistfont, 0);

  /* next two are globals */
  if ((serverlabelfont =
      (XtArgVal)XLoadQueryFont(server_display,"9x15")) != (XtArgVal)0)
    XtVaSetValues(labelw, XtNfont, serverlabelfont, 0);

  if ((serverbuttonfont =
      (XtArgVal)XLoadQueryFont(server_display,"6x9")) != (XtArgVal)0)
    XtVaSetValues(exitw, XtNfont, serverbuttonfont, 0);

  /* make sure the shell widget is wide enough */
  XawListChange(listw, list, connections + 1, 0, True);
  return shellw;
}

/*  */
/**********************************************************************
 * this is to catch those persons wicked enough to send a delete or
 * a kill to the window.  
 */
static int FatalError(disp)
  Display *disp;
{
  int i;
  if (disp == server_display) { /* won't happen if noX, we hope :-) */
    /* they just whacked the window with the list of responses.  */
    if (connections == closed)
      exit(0);
    XtCloseDisplay(server_display);
    check_quit(FATALERROR);
    longjmp(SavedEnv, 1);
  }

  /* find which display got whacked */
  for (i = 0; i < connections; i++) {
    if (connection[i].display == disp) {
      closeup(i, VIOLENT);
    }
  }
  longjmp(SavedEnv, 1);
}

/*  */
/**********************************************************************
 * determine the appropriate string for the sender of the message
 */
static void get_sender()
{
  char *username;
  char *hostname;
  struct passwd *entry;
  char header[STRLEN];

  /* get ready... */
  bzero(sender, sizeof(sender));
  bzero(header, sizeof(header));

  /* get "real" name */
  entry = getpwuid(getuid());

  if (entry) {
    username = entry -> pw_name;
    (void)sprintf(header, "Sender: %s (", entry->pw_gecos);
  }
  else {
    /* get username some other way */
    username = (char *) getlogin();
    if (!username)
      username = (char *) getenv("USER");
    (void)sprintf(header, "Sender: (");
  }

  hostname = (char *) getenv("HOST");

  if (!username && !hostname) {
    (void)sprintf(sender, "%s%s@%s)", header, "unknown", "unknown");
  }
  else if (!username) {
    (void)sprintf(sender, "%s%s@%s)", header, "unknown", hostname);
  }
  else if (!hostname) {
    (void)sprintf(sender, "%s%s@%s)", header, username, "unknown");
  }
  else {
    (void)sprintf(sender, "%s%s@%s)", header, username, hostname);
  }
  if (sender[0] == (char)0) {
    if (!silent)
      (void)fprintf(stderr, "no sender.  exiting.\n");
    exit(0);
  }
}

/*  */
/**********************************************************************
 * determine the appropriate string for the recipients
 */
static void get_recipients(recipients)
  char *recipients;
{
  int i;
  static int done = 0;
  XrmDatabase database;
  char *name;
  XrmValue namevalue;

  (void)sprintf(recipients, "%s", "Recipients: ");
  for(i = 0; i < connections; i++) {
    if (!connection[i].display)
      continue;
    if (done)
      (void)strcat(recipients, ", ");
    done = 1;
    database = NULL;
    name = NULL;
#ifndef __hpux
    if (connection[i].display->xdefaults != NULL) {
      database = connection[i].display->db;
      XrmGetResource(database, "xnotify*Name", "XtCString", 
		       &name, &namevalue);
    }
#endif
    if (name == (char *)0)
      strncpy(connection[i].name, XDisplayString(connection[i].display), 25);
    else
      strncpy(connection[i].name, namevalue.addr, 25);

    (void)strcat(recipients, connection[i].name);
  }

  if (done)
    (void)strcat(recipients, ".");
  else {
    if (!silent)
      (void)fprintf(stderr, "no recipients list.  exiting.\n");
    exit(0);
  }
}

#define WARN "\
Some connections aren't closed.\n\
Do you really want to quit?"
#define WARN2 "\
You just whacked the response window.\n\
Do you really want to quit?"

/*  */
/**********************************************************************
 * they've just asked to quit xnotify, but the connections are not
 * all acknowledged and closed.  so, we pop up a warning and ask
 * them what to do.
 */
void check_quit(which)
  int which;		/* whether this is from FatalError or dismisscmd */
{
  Widget shellw, boxw, doitw, dontdoitw;
  if (which == FATALERROR)
    if (!open_server())
      (void)fprintf(stderr, "can't open X stuff.  exiting.\n");
  shellw = XtVaCreatePopupShell("QuitShell", transientShellWidgetClass,
  				server_toplevel,
				XtNallowShellResize, True,
				XtNborderWidth, 1,
				NULL);

  boxw = XtCreateManagedWidget("box_widget", boxWidgetClass,
				shellw, NULL, 0);
  (void)XtVaCreateManagedWidget("ResponseLabel", labelWidgetClass,
				boxw,
				XtNlabel, 
				 (which == DISMISSCMD) 
				   ? (XtArgVal) WARN
				   : (XtArgVal) WARN2,
				NULL);
  doitw = XtVaCreateManagedWidget("DoQuit", commandWidgetClass,
				boxw,
				XtNlabel, (XtArgVal)"Quit anyway",
				XtNshapeStyle, XmuShapeOval,
				NULL);
  dontdoitw = XtVaCreateManagedWidget("DontQuit", commandWidgetClass,
				boxw,
				XtNlabel, (XtArgVal)"Don't Quit",
				XtNshapeStyle, XmuShapeOval,
				NULL);
  XtAddCallback(doitw, XtNcallback, dismisscmd, (XtPointer)DOQUIT);
  if (which == DISMISSCMD)
    XtAddCallback(dontdoitw, XtNcallback, dismisscmd, (XtPointer)DONTQUIT);
  else
    XtAddCallback(dontdoitw, XtNcallback, dismisscmd, (XtPointer)WHACKED);
  centerPopupOverCursor(shellw);
  XtPopup(shellw, XtGrabExclusive);
}

/*  */
/**********************************************************************
 * i originally got this idea from some of the xrn code, so credit
 * should be given to its authors.  this code is an altered variant
 * of what i saw there.
 */
void centerPopupOverCursor(widget)
  Widget widget;
{
  Window root, child;
  Dimension width, height;
  int x, y, dummy;
  unsigned int mask;

  XtRealizeWidget(widget);
  (void) XQueryPointer(XtDisplay(widget), XtWindow(widget),
                       &root, &child, &x, &y, &dummy, &dummy, &mask);
  XtVaGetValues(widget, XtNwidth,&width, XtNheight,&height, NULL);

  x -= ((int) width) / 2;
  y -= ((int) height) / 2;
  if (x + (int) width > WidthOfScreen(XtScreen(widget))) {
    x = WidthOfScreen(XtScreen(widget)) - (int) width;
  }
  if (y + (int) height > HeightOfScreen(XtScreen(widget))) {
    y = HeightOfScreen(XtScreen(widget)) - (int) height;
  }
  if (x < 0) x = 0;
  if (y < 0) y = 0;

  XtVaSetValues(widget, XtNx,(Position)x, XtNy,(Position)y, NULL);
}

/*  */
/**********************************************************************
 * this tries to open up the server display for the response window and
 * create the server display toplevel shell.  need to think about the
 * error conditions more.
 */
static int open_server()
{
  char *env_display = NULL;
  Cardinal blah = 0;

  /* response window display: try the DISPLAY environment variable first */
  env_display = (char *) getenv("DISPLAY");
  if ((server_display = XtOpenDisplay(app_context,
                        (env_display == NULL)
                          ? (String)"unix:0.0"
                          : (String)env_display,
                        "XNotify", NULL, NULL, 0, &blah, (String *)0))==NULL){
      if (!silent)
        (void)fprintf(stderr,
                      "can't open a display for the server: using stdout\n");
      print_pendings();
      useX = 0;
      nobeep = 1;
      return FAIL;
  }

  if ((server_toplevel = XtAppCreateShell("xnotify", "XNotify",
                                applicationShellWidgetClass,
                                server_display, NULL, 0)) == (Widget)0) {
    return FAIL;
  }

  return SUCCESS;
}

/*  */
/**********************************************************************
 * got to have this when i can't pass it as client_data (e.g. already used)
 */
int get_index(thisdisplay)
  Display *thisdisplay;
{
  int i;

  for (i = 0; i < connections; i++) {
    if ((connection[i].display) == thisdisplay)
      return i;
  }
  return -1;
}

/*  */
/**********************************************************************
 * print out who's pending since we couldn't open a response window
 */
static void print_pendings()
{
  int i;
  for (i = 0; i < connections; i++)
    fprintf(stdout, "%s\n", list[i]);
}

/*  */
/**********************************************************************
 * initialize the command line argument flags
 */
static void init_args(argc, argv)
  int argc;
  char **argv;
{
  int i;

  if (argc == 1) {
      (void)fprintf(stderr,"Usage: %s %s\n", argv[0], USAGE);
    exit(0);
  }
  /*
   * parse the command line, set right buttons,
   * decide which connections to try, etc. etc.
   */
  for (i = 0; i < argc - 1; i++) {
    if ((strcmp(argv[i+1], "-yes") == 0) ||
	(strcmp(argv[i+1], "-y") == 0))
      yesbutton = 1; 
    else if ((strcmp(argv[i+1], "-no") == 0) ||
	     (strcmp(argv[i+1], "-n") == 0))
      nobutton = 1; 
    else if ((strcmp(argv[i+1], "-maybe") == 0) || 
	     (strcmp(argv[i+1], "-m") == 0))
      maybebutton = 1; 
    else if ((strcmp(argv[i+1], "-ack") == 0) ||
	     (strcmp(argv[i+1], "-a") == 0))
      ackbutton = 1; 
    else if ((strcmp(argv[i+1], "-respond") == 0) ||
	     (strcmp(argv[i+1], "-r") == 0))
      respondbutton = 1; 
    else if ((strcmp(argv[i+1], "-silent") == 0) ||
	     (strcmp(argv[i+1], "-s") == 0))
      silent = 1; 
    else if (strcmp(argv[i+1], "-nobeep") == 0) 
      nobeep = 1; 
    else if (strcmp(argv[i+1], "-noX") == 0) {
      useX = 0; 
      nobeep = 1;		/* can't use XBell without X */
    }
    else if (strcmp(argv[i+1], "-v") == 0) {
      (void)printf("xnotify version %d, patchlevel %d.\n", VERSION, PATCHLEVEL);
      exit(0);
    }
    else {
      if (strchr(argv[i+1], (int)':') == NULL) {
	 (void)fprintf(stderr, "%s %s\n",
		argv[i+1], "is not a valid display.  Ignoring it.");
      }
      else {
        /* better be a display */
        connection[connections].to_check = i+1;
        connections++;
      }
    }
  }
  if (connections == 0) {
    (void)fprintf(stderr,"Usage: %s %s\n", argv[0], USAGE);
    exit(0);
  }
}

/*  */
/**********************************************************************
 * build the list to be displayed in the list widget on the response
 * window
 */
static void buildlist(i,argv)
  int i;
  char **argv;
{
  if (useX)
    list[i] = (char *) malloc((unsigned)STRLEN);
  if (!connection[i].display) {
    closed++;
    if (useX)
      (void)sprintf(list[i], "%s:  Connection never opened.",
		      argv[connection[i].to_check]);
    else
        (void)fprintf(stdout, "%s: Connection never opened.\n",
		      argv[connection[i].to_check]);
    return;
  }

  if (useX) {
    if (connection[i].name != NULL)
      (void)sprintf(list[i], "%s:  Response pending.", connection[i].name);
    else
      (void)sprintf(list[i], "%s:  Response pending.", 
		    argv[connection[i].to_check]);
  }
  else {
    if (connection[i].name != NULL)
      (void)fprintf(stdout, "%s:  Response pending.\n", connection[i].name);
    else
      (void)fprintf(stdout, "%s:  Response pending.\n", 
		    argv[connection[i].to_check]);
  }
}

/*  */
/**********************************************************************
 * get X going and open connections to everyone we plan to talk to
 */
static void start_X_and_conns(argc, argv)
  int argc;
  char **argv;
{
  int i;
  int failed = 0;			/* how many connections failed */

  XtToolkitInitialize();
  XSetIOErrorHandler(FatalError);
  app_context = XtCreateApplicationContext();

  /* initialize all the displays i'm supposed to talk to */
  for (i = 0; i < connections; i++) {
    if ((connection[i].display = 
	   XtOpenDisplay(app_context, (String)argv[connection[i].to_check],
	      "XNotify", NULL, NULL, 0, &argc, argv)) == (Display *)0) {
      failed++;
    }
  }
  /* don't REALLY need this, but why go on if i didn't connect? */
  if (failed == connections) {
    (void)fprintf(stderr, "no connections opened.  exiting.\n");
    exit(0);
  }

}

/*  */
/**********************************************************************
 * get string for when the message was sent.
 */
static void time_sent()
{
  long nowtime = 0;

  nowtime = time((long *)0);
  when = ctime(&nowtime);
  when[(strlen(when)) - 1] = '\0';
}

/*  */
/**********************************************************************
 * get the user's message (which also sets response_mess to be used
 * in the response window)
 */
static void get_message(message)
  char *message;
{
  /* get the message to send */
  if (!silent)
    (void)printf("%s\n%s\n",
	"Type message to send. End it with",
	"Control-D on a line by itself:");
  bzero(message, sizeof(message));
  (void) fread(message, sizeof(char), MESS_LEN, stdin);

  /* this will be in the response window */
  response_mess = malloc((unsigned) (64 + strlen(message)));
  message[strlen(message) - 1] = '\0';
  (void)sprintf(response_mess, "Message sent:\n\n\"%s\"\n\n", message);
}

/*  */
/**********************************************************************
 * initialize the toplevel widget and shell widget for a connection.
 * return the shell widget
 */
static Widget top_and_shell(i)
  int i;
{
  Widget shellw = NULL;

  if (connection[i].display == NULL)
    return (Widget)0;

  connection[i].toplevel = 
    XtAppCreateShell("xnotify", "XNotify", 
			applicationShellWidgetClass, 
			connection[i].display, NULL, (Cardinal) 0);

    shellw = 
      XtVaCreatePopupShell((String)sender, topLevelShellWidgetClass,
			connection[i].toplevel, XtNallowShellResize,
			True, NULL);
   if ((shellw == NULL) || (connection[i].toplevel == NULL))
     return (Widget)0;
   return shellw;
}

/*  */
/**********************************************************************
 * clean up
 */
static void clear_conn(i)
  int i;
{
  connection[i].display = NULL;
  connection[i].toplevel = NULL;
}

/*  */
/**********************************************************************
 * print (to the list or stdout) the message a user sent back.
 */
static void printout(index, message)
  int index;
  char *message;
{
  if (useX)
    (void)sprintf(list[index], "%s %s.", connection[index].name, message);
  else
    (void)fprintf(stdout, "%s %s.\n", connection[index].name, message);
}

